-- vim: ts=4 sw=4 ai et
local NAME = "Ammo-matique"
BINDING_HEADER_AMMO_MATIQUE = NAME


-- Hide under namespace Ammo
if not Ammo then Ammo = {} end
Ammo.addon = LibStub ("AceAddon-3.0"):NewAddon (NAME, "AceConsole-3.0", "AceEvent-3.0")
Ammo.status = {}

-- Localization table
local L = Ammo.strings

local RED = "|cffff0000"
local GRAY = "|cffc8c8c8"
local ORANGE = "|cffff8000"
local YELLOW = "|cffffff00"
local VOICEPATH = "Interface\\AddOns\\Ammo-matique\\Voices\\"

local function is_ammo (item)
    if not item then return end
    local _, _, _, _, _, _, _, _, slottype = GetItemInfo (item)
    return slottype == "INVTYPE_AMMO"
end

local function same_ammo_type (a, b)
    if not a or not b then return nil end
    return bit.band (a, 0x3) == bit.band (b, 0x3)
end

local function link2item (link)
    if not link then return nil end
    local _, _, item = string.find (link, "^.*Hitem:([^:]+).*")
    return item
end

local function available_ammo (ammo_type)
    local ammos = {}
    for bag = 0, 4 do
        for slot = 1, GetContainerNumSlots (bag) do
            local link = GetContainerItemLink (bag, slot)
            if is_ammo (link) and same_ammo_type (ammo_type, GetItemFamily (link)) then
                local item = link2item (link)
                -- It's retarded to even have to check for this, but we have to :(
                if GetItemCount (item) > 0 then
                    ammos[item] = link
                end
            end
        end
    end
    return ammos
end

local function effective_ilvl (ilvl, quality)
    return ilvl + (quality*9) -- This should work well for all known ammo types
end

local function cmp_by_effective_ilvl (a, b)
    local _, _, q, i = GetItemInfo (a)
    local alvl = effective_ilvl (i, q)
    _, _, q, i = GetItemInfo (b)
    local blvl = effective_ilvl (i, q)
    return alvl < blvl
end

local function find_premium_and_lowgrade ()
    local avail = available_ammo (Ammo.status.wanted_ammo_type)
    local best = {}
    local worst = {}
    for item, link in pairs (avail) do
        local _, _, quality, ilvl = GetItemInfo (item)
        ilvl = effective_ilvl (ilvl, quality)
--        Ammo.addon:Print ("debug: "..link.." calculated ilvl "..ilvl)
        if not best.ilvl or ilvl > best.ilvl then
            best.ilvl = ilvl
            best.item = item
            best.link = link
        end
        if not worst.ilvl or ilvl < worst.ilvl then
            worst.ilvl = ilvl
            worst.item = item
            worst.link = link
        end
    end
    return best, worst
end

-- Equipping related stuff
--

function Ammo.addon:RequestAmmo (ammotype)
    local premium = (ammotype == "premium")
    if premium or not Ammo.status.downgrade_disallowed then
        Ammo.status.last_equip_request_was_for_premium = premium
    end
    self:EquipSuitableAmmo ()
end


-- To avoid attempting to equip ammo multiple times, we disable listening
-- for bag updates until we get an inventory update
function Ammo.addon:DisableBagUpdates ()
    self:UnregisterEvent ("BAG_UPDATE")
    Ammo.status.register_for_bag_updates = true
end

function Ammo.addon:EnableBagUpdates ()
    Ammo.status.register_for_bag_updates = false
    self:RegisterEvent ("BAG_UPDATE")
end

function Ammo.addon:EquipAmmo (want_premium)
    local premium, lowgrade = find_premium_and_lowgrade ()
    if not premium.ilvl then
        -- No ammo available
        return false
    end
    local selected, current
    current = GetInventoryItemLink ("player", self.ammoslot)
    if want_premium then
        selected = premium
    else
        selected = lowgrade
    end
    -- If we already have the selected ammo equipped, don't re-equip it
    if (selected.item ~= link2item (current)) then
        -- Only disable bag updates if we're about to actually equip something,
        -- otherwise we'll never notice when we do get ammo
        self:DisableBagUpdates ()
        if Ammo.db.profile.general_notify then
            self:Print (L["equipping "]..selected.link)
        end
        EquipItemByName (selected.item)
        -- If we equip something, we want to warn again if we hit thresholds etc
        self:ResetWarnings ()
    end
    return true
end


function Ammo.addon:UpdateUsedAmmoType ()
    local rangedslot = GetInventorySlotInfo ("RangedSlot")
    local rangedwpn = GetInventoryItemLink ("player", rangedslot) or "" -- make GetItemInfo happy if no ranged weapon equipped
    local _, _, _, _, _, type, subtype = GetItemInfo (rangedwpn)
    local ammotype = 0
    if type == "Weapon" then
        if subtype == "Guns" then ammotype = 2
        elseif subtype == "Bows" then ammotype = 1
        elseif subtype == "Crossbows" then ammotype = 1
        end
    end
    if ammotype > 0 then
        if ammotype ~= (Ammo.status.wanted_ammo_type or 0) then
            local t = { [1] = L["arrows"], [2] = L["bullets"] }
            if Ammo.db.profile.general_notify and Ammo.db.profile.general_reload then
                self:Print (L["auto-reload active - using "]..t[ammotype])
            end
            Ammo.status.wanted_ammo_type = ammotype
        end
    else
        -- Notify if going from having auto-reload active to inactive
        if Ammo.status.wanted_ammo_type then
            if Ammo.db.profile.general_notify and Ammo.db.profile.general_reload then
                self:Print (YELLOW..L["auto-reload inactive"])
            end
            Ammo.status.wanted_ammo_type = nil
        end
    end
end


--------------------------------------------------------------
-- Hooks for keybinds, and generally useful abstractions
--

function Ammo.addon:ToggleReload ()
    local state = not Ammo.db.profile.general_reload
    Ammo.db.profile.general_reload = state
    local onoff
    if state then onoff = L["on"] else onoff = L["OFF"] end
    if Ammo.db.profile.general_notify then
        self:Print (L["auto-reload now "]..onoff)
    end
end

function Ammo.addon:ToggleSwitch ()
    local state = not Ammo.db.profile.general_switch
    Ammo.db.profile.general_switch = state
    local onoff
    if state then onoff = L["on"] else onoff = L["OFF"] end
    if Ammo.db.profile.general_notify then
        self:Print (L["auto-switch now "]..onoff)
    end
end

function Ammo.addon:EquipPremium ()
    return self:EquipAmmo (true)
end

function Ammo.addon:EquipLowgrade ()
    return self:EquipAmmo (false)
end


function Ammo.addon:EquipSuitableAmmo ()
    local ret
    if Ammo.status.last_equip_request_was_for_premium then
        ret = self:EquipPremium ()
    else
        ret = self:EquipLowgrade ()
    end
    return ret
end

function Ammo.addon:UpdateCurrentAmmoCount ()
    if GetInventoryItemLink ("player", self.ammoslot) then
        Ammo.status.ammo_count = GetInventoryItemCount ("player", self.ammoslot)
    else
        Ammo.status.ammo_count = 0
    end
end


-------------------------------------
-- Work horse for tracking ammo
--

function Ammo.addon:CheckReloadTriggers ()
    -- If auto-reload isn't enabled, don't do anything
    if not Ammo.db.profile.general_reload then return end

    -- If auto-reload isn't available, don't warn about anything
    if not Ammo.status.wanted_ammo_type then return end

    if Ammo.status.ammo_count == 0 then
        local last_ammo = GetInventoryItemLink ("player", self.ammoslot)
        if last_ammo then
            -- If we only unequipped the ammo, don't fire the out-of-current-ammo warning
            if GetItemCount (last_ammo) == 0 then
                self:OnOutOfCurrentAmmo ()
            end
        end
        if not self:EquipSuitableAmmo () then
            self:OnOutOfAllAmmo ()
        end
    elseif Ammo.status.ammo_count < Ammo.db.profile.current_low_threshold then
        self:OnLowCurrentAmmo ()
    elseif Ammo.status.ammo_count > Ammo.db.profile.current_low_threshold then
        self:ResetWarnings ()
    end
end


-------------------------------------
-- Ammo switching
--

function Ammo.addon:CheckSwitching ()
    if not Ammo.db.profile.general_switch then return end
    if not UnitCanAttack ("player", "target") then return end

    if UnitPlayerControlled ("target") then
        -- Handle PvP switching
        local isArena = IsActiveBattlefieldArena ()
        local isBG = UnitInBattleground ("player")
        if isArena then
            if Ammo.db.profile.arena_enabled then
                self:RequestAmmo (Ammo.db.profile.arena_ammotype)
            end
        elseif isBG then
            if Ammo.db.profile.bg_enabled then
                self:RequestAmmo (Ammo.db.profile.bg_ammotype)
            end
        else -- World PvP
            if Ammo.db.profile.world_enabled then
                self:RequestAmmo (Ammo.db.profile.world_ammotype)
            end
        end
    else
        -- Handle PvE switching
        local mobtype = UnitClassification ("target")
        if mobtype == "worldboss" then
            if Ammo.db.profile.bosses_enabled then
                self:RequestAmmo (Ammo.db.profile.bosses_ammotype)
            end
        elseif mobtype == "elite" or mobtype == "rareelite" then
            if Ammo.db.profile.elites_enabled then
                self:RequestAmmo (Ammo.db.profile.elites_ammotype)
            end
        else
            if Ammo.db.profile.nonelites_enabled then
                self:RequestAmmo (Ammo.db.profile.nonelites_ammotype)
            end
        end
    end
end


-------------------------------------------
-- Event control
--

function Ammo.addon:EventsOn ()
    self:RegisterEvent ("BAG_UPDATE")
    self:RegisterEvent ("UNIT_INVENTORY_CHANGED")
    self:RegisterEvent ("PLAYER_TARGET_CHANGED")
    self:RegisterEvent ("PLAYER_REGEN_DISABLED")
    self:RegisterEvent ("PLAYER_REGEN_ENABLED")
end

function Ammo.addon:EventsOff ()
    self:UnregisterEvent ("BAG_UPDATE")
    self:UnregisterEvent ("UNIT_INVENTORY_CHANGED")
    self:UnregisterEvent ("PLAYER_TARGET_CHANGED")
    self:UnregisterEvent ("PLAYER_REGEN_DISABLED")
    self:UnregisterEvent ("PLAYER_REGEN_ENABLED")
end

function Ammo.addon:ResetWarnings ()
    Ammo.status.warned_out_of_ammo = false
    Ammo.status.warned_out_of_current_ammo = false
    Ammo.status.warned_current_low = false
end

---------------------------------------------
-- Events
--

function Ammo.addon:BAG_UPDATE ()
    self:UpdateUsedAmmoType ()
    self:UpdateCurrentAmmoCount ()
    self:CheckReloadTriggers ()
end

function Ammo.addon:UNIT_INVENTORY_CHANGED (event, unit)
    if unit == "player" then
        if Ammo.status.register_for_bag_updates then
            self:EnableBagUpdates ()
        end
        self:UpdateUsedAmmoType ()
        self:UpdateCurrentAmmoCount ()
        self:CheckReloadTriggers ()
    end
end

function Ammo.addon:PLAYER_ENTERING_WORLD ()
    self:ResetWarnings ()
    self:EventsOn ()
end

function Ammo.addon:PLAYER_LEAVING_WORLD ()
    self:EventsOff ()
end

function Ammo.addon:PLAYER_TARGET_CHANGED ()
    self:CheckSwitching ()
end

function Ammo.addon:PLAYER_REGEN_DISABLED ()
    Ammo.status.downgrade_disallowed = true
    self:CheckSwitching ()
end

function Ammo.addon:PLAYER_REGEN_ENABLED ()
    Ammo.status.downgrade_disallowed = false
    self:CheckSwitching ()
end

----------------------------------------------------
-- Ammo warnings
--

function Ammo.addon:OnOutOfCurrentAmmo ()
    if not Ammo.status.warned_out_of_current_ammo then
        Ammo.status.warned_out_of_current_ammo = true
        local current_ammo = GetInventoryItemLink ("player", self.ammoslot) or (RED.."ammo")
        if Ammo.db.profile.current_exhausted_enabled then
            local msg = ORANGE..L["Warning: Out of "].."|r"..current_ammo.."|r"
            if Ammo.db.profile.current_exhausted_text then
                self:Print (msg)
            end
            if Ammo.db.profile.current_exhausted_warn then
                UIErrorsFrame:AddMessage (msg)
            end
            if Ammo.db.profile.current_exhausted_voice then
                PlaySoundFile (VOICEPATH.."cae_"..Ammo.db.profile.warnings_voiceset..".mp3")
            end
        end
    end
end

function Ammo.addon:OnOutOfAllAmmo ()
    if not Ammo.status.warned_out_of_ammo then
        Ammo.status.warned_out_of_ammo = true
        if Ammo.db.profile.all_exhausted_enabled then
            local msg = RED..L["CRITICAL: Out of all ammo!"].."|r"
            if Ammo.db.profile.all_exhausted_text then
                self:Print (msg)
            end
            if Ammo.db.profile.all_exhausted_warn then
                UIErrorsFrame:AddMessage (msg)
            end
            if Ammo.db.profile.all_exhausted_voice then
                PlaySoundFile (VOICEPATH.."aae_"..Ammo.db.profile.warnings_voiceset..".mp3")
            end
        end
    end
end

function Ammo.addon:OnLowCurrentAmmo ()
    if not Ammo.status.warned_current_low then
        Ammo.status.warned_current_low = true
        if Ammo.db.profile.current_low_enabled then
            local msg = YELLOW..L["Warning: Running low on "].."|r"..GetInventoryItemLink("player",self.ammoslot)
            if Ammo.db.profile.current_low_text then
                self:Print (msg)
            end
            if Ammo.db.profile.current_low_warn then
                UIErrorsFrame:AddMessage (msg)
            end
            if Ammo.db.profile.current_low_voice then
                PlaySoundFile (VOICEPATH.."cal_"..Ammo.db.profile.warnings_voiceset..".mp3")
            end
        end
    end
end

----------------------------------------
-- Ammo status
--

function Ammo.addon:PrintAmmoStatus ()
    if not Ammo.status.wanted_ammo_type then
        self:Print (RED..L["Don't yet know which ammo type to use - equip a ranged weapon first"])
        return
    end
    local current = GetInventoryItemLink ("player", self.ammoslot)
    local have_ammo = false
    local ammo_array = {}
    for itemid in pairs (available_ammo (Ammo.status.wanted_ammo_type)) do 
        have_ammo = true
        table.insert (ammo_array, itemid)
    end
    table.sort (ammo_array, cmp_by_effective_ilvl)
    if have_ammo then
        self:Print (L["Available ammo (in ascending order):"])
        for i, itemid in ipairs (ammo_array) do
            local _, link = GetItemInfo (itemid)
            local msg = "  "..link..GRAY.." x|r ".. GetItemCount (itemid)
            if link == current then
                msg = msg .. L[" (equipped)"]
            end
            self:Print (msg)
        end
    else
        self:Print (YELLOW.."  "..L["No ammo available!"])
    end
end

-----------------------------------------------------
-- Console command handler
--

function Ammo.addon:OnChatCmd (arg)
    if arg == "status" then
        self:PrintAmmoStatus ()
    elseif arg == "premium" then
        self:EquipPremium ()
    elseif arg == "lowgrade" then
        self:EquipLowgrade ()
    elseif arg == "config" then
        -- force expand our settings
        InterfaceOptionsFrame_OpenToCategory (Ammo.cfgdlg.BlizOptions[NAME.."\001warnings"].frame)
        InterfaceOptionsFrame_OpenToCategory (Ammo.cfgdlg.BlizOptions[NAME.."\001general"].frame)
    elseif arg == "" then
        self:Print (L["Available sub commands:"])
        self:Print ("  config - ".. L["open the configuration screen"])
        self:Print ("  status - "..L["prints ammo status"])
        self:Print ("  premium - "..L["equip premium ammo"])
        self:Print ("  lowgrade - "..L["equip lowgrade ammo"])
    else
        self:Print (RED..L["Unknown command"].." '"..arg.."'")
    end
end

--------------------------------------
-- Initialisation
--

function Ammo.addon:OnEnable ()
    Ammo.db = LibStub ("AceDB-3.0"):New ("AmmomatiqueDB", Ammo.defaults)
    LibStub ("AceConfigRegistry-3.0"):RegisterOptionsTable (NAME, Ammo.config)
    Ammo.config.args.profiles = LibStub ("AceDBOptions-3.0"):GetOptionsTable (Ammo.db)
    
    Ammo.cfgdlg = LibStub ("AceConfigDialog-3.0")
    Ammo.cfgdlg:AddToBlizOptions (NAME, nil, nil, "general")
    Ammo.cfgdlg:AddToBlizOptions (NAME, L["Ammo warnings"], NAME, "warnings")
    Ammo.cfgdlg:AddToBlizOptions (NAME, L["PvE ammo"], NAME, "pve")
    Ammo.cfgdlg:AddToBlizOptions (NAME, L["PvP ammo"], NAME, "pvp")
    Ammo.cfgdlg:AddToBlizOptions (NAME, L["Profiles"], NAME, "profiles")

    self.ammoslot = GetInventorySlotInfo ("AmmoSlot")
    self:UpdateUsedAmmoType ()
    self:UpdateCurrentAmmoCount ()
    if not Ammo.status.wanted_ammo_type then
        if Ammo.db.profile.notify and Ammo.db.profile.general_reload then
            self:Print (RED..L["no ranged weapon currently equipped, auto-reload inactive until one has been equipped"])
        end
    else
    end
    self:RegisterEvent ("PLAYER_ENTERING_WORLD")
    self:RegisterEvent ("PLAYER_LEAVING_WORLD")
    self:RegisterChatCommand ("ammo", "OnChatCmd")
end


-------------------------------------
-- Override the colour from Print()
local mt = getmetatable (Ammo.addon) or {}
mt["__tostring"] = function () return "|cff00cc00"..NAME.."|r" end
setmetatable (Ammo.addon, mt)

